﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.ComponentModel;
using System.Collections.ObjectModel;
using System.Windows.Data;
using System.Windows.Input;
using System.Diagnostics;
using System.Windows;
using System.Threading;

namespace AsyncExecution
{
    public class ViewModel : INotifyPropertyChanged
    {
        public ViewModel()
        {
            //Create Collection and View
            PersonCollection = new ObservableCollection<Person>();
            PersonsView = new ListCollectionView(PersonCollection);

            //Command Bindings
            FindPersons = new CommandBinding(ApplicationCommands.Find, FindPersons_Executed);
            GetAges = new CommandBinding(ApplicationCommands.Open, GetAges_Executed);

            //Create Data Provider
            provider = new DataProvider();

            //Instantiate the AsyncExecutor
            AsyncExecutor = new AsyncExecutor();

            StatusMessage = "Initialization done";
        }

        /// <summary>
        /// Gets or sets the async executor.
        /// A property is used here to be able to make the UI bind to the IsBusy property 
        /// of the AsyncExecutor
        /// </summary>
        /// <value>The async executor.</value>
        public AsyncExecutor AsyncExecutor { get; private set; }

        private DataProvider provider = null;

        private ObservableCollection<Person> PersonCollection { get; set; }

        /// <summary>
        /// View - will be bound to the ListView.
        /// </summary>
        /// <value>The view</value>
        public ListCollectionView PersonsView { get; set; }


        #region CommandBinding: FindPersons
        public CommandBinding FindPersons { get; set; }

        private void FindPersons_Executed(object sender, ExecutedRoutedEventArgs args)
        {
            if (!IsUseAsyncMode)
            {
                //Synchronous execution
                GetPersonsSynchronously();
            }
            else
            {
                //Uncomment to use the background worker
                //GetPersonsAsync_Using_BackgroundWorker();

                //Uncomment to use the much simpler AsyncExecutor
                AsyncExecutor.StartExecution(GetPersonsAsync);
            }
        }

        #endregion


        private void GetPersonsSynchronously()
        {
            PersonCollection.Clear();
            StatusMessage = "Loading Persons...";

            List<string> persons = provider.GetPersons();

            foreach (var name in persons)
            {
                PersonCollection.Add(new Person { Name = name });
            }

            StatusMessage = "Finished loading persons";


            MessageBoxResult result = MessageBox.Show("Do you want me to look for the age of the persons?", "Get Age?", MessageBoxButton.OKCancel, MessageBoxImage.Question);

            if (result == MessageBoxResult.Cancel)
            {
                return;
            }

            DoGetAgesSynchronously();
        }

        #region GetPersons - using an BackgroundWorker
        private BackgroundWorker getPersonsWorker = null;

        /// <summary>
        /// Start the background worker to get the persons - QUITE CUMBERSOME!
        /// </summary>
        private void GetPersonsAsync_Using_BackgroundWorker()
        {
            PersonCollection.Clear();
            StatusMessage = "Loading Persons...";

            if (getPersonsWorker == null)
            {
                getPersonsWorker = new BackgroundWorker();

                //Attach to the events
                getPersonsWorker.DoWork += new DoWorkEventHandler(getPersonsWorker_DoWork);
                getPersonsWorker.RunWorkerCompleted += new RunWorkerCompletedEventHandler(getPersonsWorker_RunWorkerCompleted);
            }

            //Start the worker
            getPersonsWorker.RunWorkerAsync();
        }

        private void getPersonsWorker_DoWork(object sender, DoWorkEventArgs e)
        {
            List<string> persons = provider.GetPersons();

            e.Result = persons;
        }

        private void getPersonsWorker_RunWorkerCompleted(object sender, RunWorkerCompletedEventArgs e)
        {
            List<string> persons = e.Result as List<string>;

            PersonCollection.Clear();

            foreach (var name in persons)
            {
                PersonCollection.Add(new Person { Name = name });
            }

            StatusMessage = "Finished loading persons";
        }
        #endregion

        /// <summary>
        /// Gets the persons async using a MUCH SIMPLER way to do it.
        /// </summary>
        /// <returns></returns>
        private IEnumerator<SwitchExecutionContext> GetPersonsAsync()
        {
            yield return SwitchExecutionContext.ToUIThread;
            PersonCollection.Clear();
            StatusMessage = "Loading Persons...";

            yield return SwitchExecutionContext.ToWorkerThread;
            List<string> persons = provider.GetPersons();

            yield return SwitchExecutionContext.ToUIThread;
            foreach (var name in persons)
            {
                PersonCollection.Add(new Person { Name = name });
            }

            StatusMessage = "Finished loading persons";

            MessageBoxResult result = MessageBox.Show("Do you want me to look for the age of the persons?", "Get Age?", MessageBoxButton.OKCancel, MessageBoxImage.Question);

            if (result == MessageBoxResult.Cancel)
            {
                yield break;
            }

            //Switch to the worker thread before calling the execute method
            yield return SwitchExecutionContext.ToWorkerThread;
            AsyncExecutor.Execute(DoGetAgesAsynchronously);
        }




        #region CommandBinding: GetAges
        public CommandBinding GetAges { get; set; }

        private void GetAges_Executed(object sender, ExecutedRoutedEventArgs args)
        {
            if (!IsUseAsyncMode)
            {
                //Synchronous execution:
                DoGetAgesSynchronously();
            }
            else
            {
                //Better: Start the Async execution using the AsyncExecutor
                AsyncExecutor.StartExecution(DoGetAgesAsynchronously);
            }
        }

        #endregion

        private void DoGetAgesSynchronously()
        {
            foreach (Person person in PersonsView)
            {
                PersonsView.MoveCurrentTo(person);

                StatusMessage = "Getting age for: " + person.Name;

                int age = provider.GetAge(person.Name);

                person.Age = age;
            }

            StatusMessage = "All ages identified";
        }

        private IEnumerator<SwitchExecutionContext> DoGetAgesAsynchronously()
        {
            foreach (Person person in PersonsView)
            {
                yield return SwitchExecutionContext.ToUIThread;
                PersonsView.MoveCurrentTo(person);

                StatusMessage = "Getting age for: " + person.Name;

                yield return SwitchExecutionContext.ToWorkerThread;
                int age = provider.GetAge(person.Name);

                yield return SwitchExecutionContext.ToUIThread;
                person.Age = age;
            }

            StatusMessage = "All ages identified";
        }

        /// <summary>
        /// Whether to use the Synchronous or Asynchronous mode
        /// </summary>
        public bool IsUseAsyncMode { get; set; }


        #region Property: string StatusMessage
        /// <summary>
        /// Holder for StatusMessage
        /// </summary>
        private string statusMessage;

        /// <summary>
        /// Get and set StatusMessage
        /// </summary>
        public string StatusMessage
        {
            get { return statusMessage; }
            set
            {
                statusMessage = value;
                FirePropertyChanged("StatusMessage");
            }
        }
        #endregion

        #region INotifyPropertyChanged Members

        private void FirePropertyChanged(string propertyName)
        {
            if (PropertyChanged != null)
                PropertyChanged(this, new PropertyChangedEventArgs(propertyName));
        }


        public event PropertyChangedEventHandler PropertyChanged;

        #endregion
    }

    public class Person : INotifyPropertyChanged
    {
        #region Property: string Name
        /// <summary>
        /// Holder for Name
        /// </summary>
        private string name;

        /// <summary>
        /// Get and set Name
        /// </summary>
        public string Name
        {
            get { return name; }
            set
            {
                name = value;
                FirePropertyChanged("Name");
            }
        }
        #endregion

        #region Property: int Age
        /// <summary>
        /// Holder for Income
        /// </summary>
        private int age;

        /// <summary>
        /// Get and set Income
        /// </summary>
        public int Age
        {
            get { return age; }
            set
            {
                age = value;
                FirePropertyChanged("Age");
            }
        }
        #endregion
        

        #region INotifyPropertyChanged Members

        private void FirePropertyChanged(string propertyName)
        {
            if (PropertyChanged != null)
                PropertyChanged(this, new PropertyChangedEventArgs(propertyName));
        }


        public event PropertyChangedEventHandler PropertyChanged;

        #endregion
    }
}

